import React from 'react'
import {expect} from 'chai'
import {mount, shallow} from 'enzyme'
import {describe, it} from 'mocha'

import ModelRunConfigForm from '../../../src/components/ConfigForm/ModelRunConfigForm.js'

import {sos_model_run, sos_models, scenarios, narratives} from '../../helpers.js'
import {empty_object, empty_array} from '../../helpers.js'

describe.skip('<ModelRunConfigForm />', () => {


})
